from army.api.log import log
from army.api.debugtools import print_stack
from army.api.project import load_project
from army.api.repository import load_repositories
from army.api.click import verbose_option 
from army.army import cli, dependencies
import click
import os

from army.army import prefix

@dependencies.command(name='repos', help='List available repositories')
@verbose_option()
@click.pass_context
def repos(ctx, **kwargs):
    log.info(f"repos")
    
    config = ctx.parent.config
        
    # build repositories list
    repositories = load_repositories(config, prefix)
    if len(repositories)==0:
        print("no repository configured")
        return 
    
    column_name = ['name']
    column_type = ['type']
    column_uri = ['uri']

    for r in repositories:
        column_name.append(r.name)
        column_type.append(r.type)
        column_uri.append(r.uri)

    max_name = len(max(column_name, key=len))
    max_type = len(max(column_type, key=len))
    max_uri = len(max(column_uri, key=len))
  
    if len(column_name)>0:
        for i in range(len(column_name)):
            print(f"{column_name[i].ljust(max_name, ' ')} | ", end='')
            print(f"{column_type[i].ljust(max_type)} | ", end='')
            print(f"{column_uri[i].ljust(max_uri)}", end='')
            print()



import logging

FORMAT = '[%(levelname)s] %(message)s'
logging.basicConfig(format=FORMAT)

log = logging.getLogger('console')
log.setLevel('ERROR') # default log mode warning

def get_log_level():
    if log.level==10:
        return "debug"
    elif log.level==20:
        return "info"
    elif log.level==30:
        return "warning"
    elif log.level==40:
        return "error"
    else:
        return "fatal"

def set_log_level(level):
    if level is None or level=='fatal':
        log.setLevel('CRITICAL')        
    elif level=='error':
        log.setLevel('ERROR')
    elif level=='warning':
        log.setLevel('WARNING')
    elif level=='info':
        log.setLevel('INFO')
    elif level=='debug':
        log.setLevel('DEBUG')
            


from army.api.log import log
from army.api.debugtools import print_stack
from army.api.project import load_project
from army.army import prefix
from army.api.click import verbose_option 
from army.api.package import load_installed_packages
from army.army import cli
import click
import sys


@cli.command(name='init', help='Create a new project')
@verbose_option()
@click.pass_context
def init(ctx, **kwargs):
    log.info(f"init")
    
    # load configuration
    config = ctx.parent.config
    project_config = None
    try:
        # load project configuration
        project_config = load_project(config)
        config = project_config
    except Exception as e:
        print_stack()
        log.debug(e)
        log.info(f"no project loaded")

    print("Not yet implemented")



from army.api.config import Config, ConfigException,  ConfigString, ConfigInt
from army.api.log import log
import os

# get current file path to find ressource files
path = os.path.dirname(__file__)
log.setLevel('DEBUG') # default log mode warning


def raised(func, *args,**kwargs):
    res = False
    try:
        func(*args,**kwargs)
    except Exception as e:
        res = type(e)
#         print(e)
    return res



s = ConfigString(value="value")
assert str(s)=="value"

# check raise with improper values
assert raised(ConfigString)==ConfigException
assert raised(ConfigString, 0)==ConfigException
assert raised(ConfigString, True)==ConfigException
assert raised(ConfigInt)==ConfigException
assert raised(ConfigInt, "aaaaa")==ConfigException

class TestConfig(Config):
    def __init__(self, value=None, parent=None):
        super(TestConfig, self).__init__(value, parent,
            fields={
                'value1': [ ConfigString, "v1" ],
                'value2': [ ConfigInt, 5 ],
                'value3': [ ConfigInt, 10 ]
            })
    

# no config set, check we are finding the default values
c1 = TestConfig() ;
assert str(c1.get('value1'))=='v1'
assert str(c1.value1)=='v1'
assert int(c1.value2)==5
assert int(c1.value3)==10
print("c1:", c1.expand())

# values are set, check that we find them 
c2 = TestConfig(value={
    "value1": "v2",
    "value2": 1
    })
assert str(c2.value1)=='v2'
assert int(c2.value2)== 1
assert int(c2.value3)== 10
print("c2:", c2.expand())

# child configuration overrides 
c3 = TestConfig(value={
    "value2": 2
    }, parent=c2)
print(type(c3.value1))
assert str(c3.value1)=='v2'
assert int(c3.get('value2'))==2
assert int(c3.get('value3'))==10
print("c3:", c3.expand())

class ConfigStringEnum(ConfigString):
    def __init__(self, value=None, parent=None):
        super(ConfigStringEnum, self).__init__(value, parent)

    def check(self):
        super(ConfigStringEnum, self).check() 
        if self.value() not in {"a", "b", "c"}:
            raise ConfigException(f"'{self.value()}': invalid value")

assert raised(ConfigStringEnum)==ConfigException
assert raised(ConfigStringEnum, "d")==ConfigException
c4 = ConfigStringEnum("a")
assert str(c4)=='a'
print("c4:", c4.expand())


class TestConfig2(Config):
    def __init__(self, value=None, parent=None):
        super(TestConfig2, self).__init__(value, parent,
            fields={
                'config1': [ TestConfig, {} ],
                'config2': [ TestConfig, { "value1": "test" } ],
                'value1': [ ConfigStringEnum, "b" ]
            })
    

c5 = TestConfig2()
print("c5:", c5.expand())
assert str(c5.get("config1").get("value1"))=="v1"
assert int(c5.get("config1").get("value2"))==5
assert str(c5.get("config2").get("value1"))=="test"
assert int(c5.get("config2").get("value2"))==5
assert str(c5.get("value1"))=="b"

c6 = TestConfig2(value={
    "config1": {"value1": 'test1'},
    "value1": "c"
    })
print("c6:", c6.expand())
assert str(c6.get("config1").get("value1"))=="test1"
assert int(c6.get("config1").get("value2"))==5
assert str(c6.get("config2").get("value1"))=="test"
assert int(c6.get("config2").get("value2"))==5
assert str(c6.get("value1"))=="c"

c7 = TestConfig2(value={
    "config1": {"value2": 10},
    }, parent=c6)
print("c7:", c7.expand())
assert str(c7.get("config1").get("value1"))=="v1"
assert int(c7.get("config1").get("value2"))==10
assert str(c7.get("config2").get("value1"))=="test"
assert int(c7.get("config2").get("value2"))==5
assert str(c7.get("value1"))=="c"

class TestConfig3(Config):
    def __init__(self, value=None, parent=None):
        super(TestConfig3, self).__init__(value, parent,
            fields={
                'value2': [ ConfigString, "test" ]
            })
c8 = TestConfig3(value={
    "value2": "test2",
    }, parent=c7)
print("c8:", c8.expand())
assert str(c8.get("config1").get("value1"))=="v1"
assert int(c8.get("config1").get("value2"))==10
assert str(c8.get("config2").get("value1"))=="test"
assert int(c8.get("config2").get("value2"))==5
assert str(c8.get("value1"))=="c"
assert str(c8.get("value2"))=="test2"

c8.set("value1", "a")
assert str(c8.get("value1"))=="a"

assert raised(c8.set, "value1", "d")==ConfigException
assert str(c8.get("value1"))=="a"


##############################
# test army config files loading
from army.api.config import ArmyConfigFile, load_configuration, ConfigRepository, ConfigRepositoryDict, ConfigRepositoryFile
from army.api.config import ConfigList

# check simple file load
c9 = ArmyConfigFile(file=os.path.join(path, "test_data/etc/army/army.toml"))
print("c9:", c9.expand())
assert str(c9.get("verbose"))=="debug"

class ConfigListItem(Config):
    def __init__(self, parent=None, value=None):
        super(ConfigListItem, self).__init__(
            parent=parent,
            value=value,
            fields={
                'name': [ ConfigString, "" ],
                'type': [ ConfigString, "" ],
                'uri': [ ConfigString, "" ],
            }
        )

c10a = ConfigList(
    value=[
            {'name': 'repo_test1', 'type': 'git-local', 'uri': '~/git/repo_test1'}, 
            {'name': 'repo_test2', 'type': 'git-local', 'uri': '~/git/repo_test2'}
        ],
    field=[ConfigListItem, {}]
    )
c10b = ConfigList(
    value=[
            {'name': 'repo_test3', 'type': 'git-local', 'uri': '~/git/repo_test3'}, 
            {'name': 'repo_test4', 'type': 'git-local', 'uri': '~/git/repo_test4'}
        ],
    field=[ConfigListItem, {}],
    parent=c10a
    )
assert len(c10a)==2
assert len(c10b)==4
assert str(c10a[0].get('name'))=='repo_test1'
assert str(c10b[0].get('name'))=='repo_test3'
assert str(c10b[2].get('name'))=='repo_test1'
i=0
for item in c10b:
    print(f"c10b[{i}]:", item) #.get('name'))
    i += 1 
print("c10b:", c10b.expand())


c11a = ConfigRepositoryDict(
    value={
            'repo_test1': {'type': 'git-local', 'uri': '~/git/repo_test1'}, 
            'repo_test2': {'type': 'git-local', 'uri': '~/git/repo_test2'},
            'repo_test3': {'type': 'git-local', 'uri': '~/git/repo_test3'}
        },
    )
c11b = ConfigRepositoryDict(
    value={
            'repo_test3': {'type': 'git', 'uri': 'http://github.com/repo_test3'}, 
            'repo_test4': {'type': 'git-local', 'uri': '~/git/repo_test4'}
        },
    parent=c11a
    )
assert len(c11a)==3
assert len(c11b)==4
assert str(c11a['repo_test1'].get('uri'))=='~/git/repo_test1'
assert str(c11a['repo_test3'].get('type'))=='git-local'
assert str(c11b['repo_test3'].get('type'))=='git'

for item in c11b:
    print(f"c11b['{item}']:", c11b[item]) #.get('name'))
print("c11b:", c11b.expand())
# 
c12 = ConfigRepositoryFile(file=os.path.join(path, "test_data/etc/army/repo.d/repo_test1-2.toml"))
assert len(c12.repo)==2
assert str(c12.get("repo")["repo_test1"].get('type'))=='git-local'
# 
print(f"load test files from {path}")
config=load_configuration(prefix=os.path.join(path, "test_data"))
print("c12:", config.expand())
assert len(config.repo)==6
assert str(config.repo['main'].type)=='git-local'
assert str(config.repo['repo_test2'].type)=='git'
# check iterator
repos={}
for repo in config.repo:
    repos[repo] = repo
print(repos)
assert len(repos)==6



# click extentions and helpers
from army.api.log import set_log_level

import click

# utility tool to add member methods to a class
from functools import wraps
def add_method(cls):
    def decorator(func):
        @wraps(func) 
        def wrapper(self, *args, **kwargs): 
            return func(self, *args, **kwargs)
        setattr(cls, func.__name__, wrapper)
        # Note we are not binding func, but wrapper which accepts self but does exactly the same as func
        return func # returning func means func can still be used normally
    return decorator


def _check_section_params(cmd):
    if not isinstance(cmd, Section) or len(cmd.params)==0:
        return
    
    raise RuntimeError(
        f"Section '{cmd.name}' can not have any parameter as they would be ignored."
    )

@add_method(click.MultiCommand)
def format_commands(self, ctx, formatter):
    """Extra format methods for multi methods that adds all the commands
    after the options.
    """
    default_section = "Commands"
    
    sections = {}
    # add default section
    sections["Commands"] = []
    for subcommand in self.list_commands(ctx):
        cmd = self.get_command(ctx, subcommand)
        # What is this, the tool lied about a command.  Ignore it
        if cmd is None:
            continue
        if cmd.hidden:
            continue
        
        if isinstance(cmd, Section):
            # add section subcommands
            sections[cmd.name] = []
            for sxnsubcommand in cmd.list_commands(ctx):
                sxncmd = cmd.get_command(ctx, sxnsubcommand)
                # What is this, the tool lied about a command.  Ignore it
                if sxncmd is None:
                    continue
                if sxncmd.hidden:
                    continue
                
                sections[cmd.name].append((sxnsubcommand, sxncmd))
        else:
            sections[default_section].append((subcommand, cmd))
    
    # allow for 3 times the default spacing
    for section in sections:
        commands = sections[section]
        
        if len(commands):
            limit = formatter.width - 6 - max(len(cmd[0]) for cmd in commands)

            rows = []
            for subcommand, cmd in commands:
                help = cmd.get_short_help_str(limit)
                rows.append((subcommand, help))

            if rows:
                with formatter.section(section):
                    formatter.write_dl(rows)


@add_method(click.MultiCommand)
def resolve_command(self, ctx, args):
    cmd_name = click.utils.make_str(args[0])
    original_cmd_name = cmd_name

    # Get the command
    cmd = self.get_command(ctx, cmd_name)
    
    if isinstance(cmd, Section):
        # in case we find a section we do not consume argument, it will be
        # consumed later by the section
        return cmd_name, cmd, args
    
    # If we can't find the command but there is a normalization
    # function available, we try with that one.
    if cmd is None and ctx.token_normalize_func is not None:
        cmd_name = ctx.token_normalize_func(cmd_name)
        cmd = self.get_command(ctx, cmd_name)

    # If we don't find the command we want to show an error message
    # to the user that it was not provided.  However, there is
    # something else we should do: if the first argument looks like
    # an option we want to kick off parsing again for arguments to
    # resolve things like --help which now should go to the main
    # place.
    if cmd is None and not ctx.resilient_parsing:
        if click.parser.split_opt(cmd_name)[0]:
            self.parse_args(ctx, ctx.args)
        ctx.fail(f"No such command '{original_cmd_name}'.")

    return cmd_name, cmd, args[1:]


@add_method(click.Group)
def section(self, *args, **kwargs):
    """A shortcut decorator for declaring and attaching a section to
    the group.  This takes the same arguments as :func:`group` but
    immediately registers the created command with this instance by
    calling into :meth:`add_command`.
    """
    from .decorators import section

    def decorator(f):
        cmd = section(*args, **kwargs)(f)
        self.add_command(cmd)
        return cmd

    return decorator

@add_method(click.Group)
def get_command(self, ctx, cmd_name):
    cmd = self.commands.get(cmd_name)
    if cmd is None:
        # look if there is any section containing the command
        for section in self.commands:
            if isinstance(self.commands[section], Section) and self.commands[section].get_command(ctx, cmd_name):
                return self.commands[section]
    return cmd

@add_method(click.Group)
def list_commands(self, ctx):
    print("---", self.commands)
    return sorted(self.commands)

class Section(click.Group):
    """A section allows a command to split subcommands into categories attached. This can be seen
    as creating an anonymous group of subcommands

    :param commands: a dictionary of commands.
    """

    def __init__(
        self,
        name=None,
        **attrs,
    ):
        super(Section, self).__init__(name, **attrs)

    def get_params(self, ctx):
        # TODO this check is too late, should appear when argument is added
        _check_section_params(self)
        return super(Section, self).get_params(ctx)

    def get_command(self, ctx, cmd_name):
        return self.commands.get(cmd_name)

    def list_commands(self, ctx):
        return sorted(self.commands)



from army.api.log import log
from army.api.debugtools import print_stack
from army.api.project import load_project
from army.api.repository import load_repositories
from army.army import prefix
from army.api.click import verbose_option 
from army.api.package import load_installed_packages, load_installed_package
from army.army import cli, dependencies
import click
import os
import sys

@dependencies.command(name='uninstall', help='uninstall package')
@verbose_option()
# @click.option('--save', help='Update project package list', is_flag=True)    # TODO
@click.argument('name', nargs=-1)
@click.pass_context
def uninstall(ctx, name, **kwargs):
    log.info(f"uninstall {name} {kwargs}")
    
    config = ctx.parent.config

    if len(name)==0:
        print("nothing to uninstall", file=sys.stderr)
        exit(1)
        
    # build repositories list
    repositories = load_repositories(config, prefix)
    
    packages = []

    for package in name:
        pkg = load_installed_package(package, prefix=prefix)
        if pkg is None:
            print(f"{package}: package not installed", file=sys.stderr)
            exit(1)
        packages.append(pkg)
    
    for package in packages:
        package.uninstall()
    


import army.plugin.package.search
import army.plugin.package.install
import army.plugin.package.repos
import army.plugin.package.list
import army.plugin.package.uninstall
import army.api.click as click
from army.army import cli



from army.api.log import log
from army.api.version import Version
from army.api.config import load_configuration, ConfigException
from army.api.project import load_project, ConfigProject, ArmyProjectFile
import os

def raised(func, *args,**kwargs):
    res = False
    try:
        func(*args,**kwargs)
    except Exception as e:
        res = type(e)
#         print(e)
    return res

# get current file path to find ressource files
path = os.path.dirname(__file__)
log.setLevel('DEBUG') # default log mode warning

print(f"load test files from {path}")
config=load_configuration(prefix=os.path.join(path, "test_data"))
print("config", config.expand())

assert raised(ConfigProject)==ConfigException

p1 = ConfigProject(value={
        'name': "project",
        'description': "test",
        'version': "1.0.0"
    })
print( p1.expand() )

p2 = ArmyProjectFile(file=os.path.join(path, "test_data/~/git/project1/army.toml"), parent=config)
assert str(p2.project.name)=="project1"



#!/bin/python3
import os
import sys
import pkg_resources
from itertools import chain

sys.path.append(os.path.dirname(pkg_resources.resource_filename(__name__, "army.py")))

from army.api.config import ArmyConfig, load_configuration
from army.api.log import log, get_log_level
from army.api.debugtools import print_stack
from army.api.project import load_project
from army.api.plugin import load_plugin
from army.api.click import verbose_option 
import army.api.click as click

# TODO add autocomplete https://kislyuk.github.io/argcomplete/

# Git coupling
# see https://git-scm.com/docs/git-worktree

# configuration:
# /etc/army        # global to all users configuration
# ~/.army          # local user configuration
# <project>/.army  # local project configuration

# type of repositories
#  - http: indexed repository with versioned tar files only
#  - remote git: allow to install pushed master/tagged version directly from git
#  - local git: allow to install dev versions with symbolic links 

# /etc/army/
#   repo.d/
#   army.toml

# ~/.army/
#   repo.d/
#   army.toml
 
# <project>/.army/
#   repo.d/
#   army.toml

# firmware project structure:
# board/    # contain supported boards
#   board-1/
#     cmake/
#       board.cmake      # board specific cmake file
#       packages.cmake   # managed by army, references installed packages
#     core/  # contains board specific code
#       board.h
#       init.cpp
#     ld/    # contains ld scripts
#     packages/  # contains army installed packages
#     board.toml  # board description and dependencies
#   board-2/
#   <...>
# firmware/  # firmware, not specific to any board
# output/    # build result
# project.toml # project description and global dependencies

# default configuration
root_config = ArmyConfig()
config = None

# create the config parser, this parser is mainly used to catch logger verbosity
@click.group(invoke_without_command=True, 
             no_args_is_help=False, 
             add_help_option=False, 
             context_settings=dict(
                 resilient_parsing=True))
@verbose_option()
@click.pass_context
def cli_init(ctx, v, **kwargs):
    # configure logger
    root_config.set("verbose", get_log_level())


# create the top-level parser
@click.group()
@verbose_option()
@click.option('-t', '--target', help='select target')
@click.pass_context
# TODO add version with version_option
def cli(ctx, target, **kwargs):
    global config
    ctx.config = config

    if target is not None:
        config.target = target

@cli.section("Dependencies Commands")
@click.pass_context
def dependencies(ctx, **kwargs):
    # recopy parent config in context
    ctx.config = ctx.parent.config

@cli.section("Build Commands", chain=True)
@click.pass_context
def build(ctx, **kwargs):
    # recopy parent config in context
    ctx.config = ctx.parent.config

# path prefix, used to provide unit tests data path
prefix = None
#prefix = os.path.join(os.path.dirname(__file__), "unit_tests/test_project_data")
#prefix = os.path.join(os.path.dirname(__file__), "unit_tests/test_compile_data")
#prefix = os.path.join(os.path.dirname(__file__), "unit_tests/test_data")

def main():
    global prefix
    global config
    
    try:
        # cli_init will initialize the logger only, everything else is ignored at this point
        # we need to load the plugins before showing any help
        cli_init()
    except:
        pass

    # load army configuration files
    if prefix is not None:
        log.debug(f"using {prefix} as path prefix")
    config = load_configuration(parent=root_config, prefix=prefix) 

    # load internal plugins
    import army.plugin.repository
    import army.plugin.package
#     import army.plugin.build

    # load plugins
    # TODO load plugins from installed packages
    if os.path.exists('army.toml'):
        try:
            project = load_project()
            for plugin in project.plugins:
                try:
                    load_plugin(plugin, config)
                except Exception as e:
                    print_stack()
                    print(f"{e}")
        except Exception as e:
            print_stack()
            print(f"army.toml: {e}", file=sys.stderr)
            exit(1)
    # parse command line
    cli() 
    
    
if __name__ == "__main__":

    main()



from army.api.log import log
from army.api.config import load_configuration
from army.api.repository import load_repositories
import os

# get current file path to find ressource files
path = os.path.dirname(__file__)
log.setLevel('DEBUG') # default log mode warning

print(f"load test files from {path}")
config=load_configuration(prefix=os.path.join(path, "test_data"))
print("config", config.expand())

# load repository plugin
import army.plugin.repository
from army.plugin.repository.local_git_repository import LocalGitRepository

r1 = LocalGitRepository(name="project1", path=os.path.join(path, "test_data/~/git/project1"))
r1.load()
assert r1.TYPE=="git-local"
assert r1.DEV==True
assert len(r1.packages())==1
print(r1.search("project1"))
print(r1.search("project1:1"))

# load test repositories
repositories = load_repositories(config)
print(repositories)


# load search plugin
import army.plugin.package
from army.plugin.package.search import SearchCommand

args = object()
args.NAME = "project1"
c1 = SearchCommand(None)
c1.execute(config, args)



from army.api.version import Version, VersionRange

class SchemaException(Exception):
    def __init__(self, message):
        self.message = message


class Schema(object):
    
    def __init__(self, data, schema={}):
        self._data = data
        self._schema = schema
        
    def check(self):
        for item in self._data:
            
            if item not in self._schema:
                raise SchemaException(f"{item}: unknown item")
            else:
                self._schema[item].check(self._data[item])

class ValidatorException(Exception):
    def __init__(self, message):
        self.message = message
                
class Validator(object):
    def __init__(self):
        pass
    
    def check(self, value):
        pass
    
class Variant(Validator):
    def __init__(self):
        super(Variant, self).__init__()
    
    def check(self, value):
        pass

class Int(Validator):
    def __init__(self):
        super(Int, self).__init__()
    
    def check(self, value):
        raise "not implemented"

class String(Validator):
    def __init__(self):
        super(String, self).__init__()
    
    def check(self, value):
        if not isinstance(value, str):
            raise ValidatorException(f"'{value}' is not a valid string")

class VersionRangeString(Validator):
    def __init__(self):
        super(VersionRangeString, self).__init__()
    
    def check(self, value):
        try:
            version = VersionRange(value)
        except Exception as e:
            raise ValidatorException(f"'{value}' is not a valid version")

class VersionString(Validator):
    def __init__(self):
        super(VersionString, self).__init__()
    
    def check(self, value):
        try:
            version = Version(value)
        except Exception as e:
            raise ValidatorException(f"'{value}' is not a valid version")

class Optional(Validator):
    def __init__(self, schema):
        super(Optional, self).__init__()
        self._schema = schema
        
    def check(self, value):
        if value is not None:
            self._schema.check(value)

class PackageString(Validator):
    def __init__(self):
        super(PackageString, self).__init__()
    
    def check(self, value):
        if not isinstance(value, str):
            raise ValidatorException(f"'{value}' is not a valid package name")
        # TODO: validate name
        
class Array(Validator):
    def __init__(self, schema):
        super(Array, self).__init__()
        self._schema = schema
    
    def check(self, values):
        if not isinstance(values, list):
            raise ValidatorException(f"'{values}' is not a valid array list")
        for value in values:
            self._schema.check(value)

class Dict(Validator):
    def __init__(self, schema):
        super(Dict, self).__init__()
        self._schema = schema
    
    def check(self, values):
        if not isinstance(values, dict):
            raise ValidatorException(f"'{values}' is not a valid dictionnary")
        for value in values:
            if value not in self._schema:
                raise SchemaException(f"{value}: unknown item")
            else:
                self._schema[value].check(values[value])

class VariableDict(Validator):
    def __init__(self, schemakey, schemavalue):
        super(VariableDict, self).__init__()
        self._schemakey = schemakey
        self._schemavalue = schemavalue
    
    def check(self, values):
        if not isinstance(values, dict):
            raise ValidatorException(f"'{values}' is not a valid dictionnary")
        for value in values:
            self._schemakey.check(value)
            self._schemavalue.check(values[value])



from army.api.repository import register_repository
from army.plugin.repository.local_git_repository import LocalGitRepository
from army.plugin.repository.git_repository import GitRepository

# init plugin

# register repository types
register_repository(LocalGitRepository)
register_repository(GitRepository)



from army.api.version import Version
from army.api.log import log
from army.api.debugtools import print_stack
from army.api.schema import Schema, String, VersionString, Optional, PackageString, Array, Dict, VariableDict, Variant, VersionRangeString
import os
import toml
import shutil

def load_installed_packages(local=True, _global=True, prefix=None):
    res = []

    def _search_dir(path):
        if os.path.exists(os.path.expanduser(path))==False:
            return []
        
        res = []
        for package in os.listdir(os.path.expanduser(path)):
            try:
                pkg = _load_installed_package(os.path.join(path, package))
                res.append(pkg)
            except Exception as e:
                print_stack()
                log.debug(e)
                log.error(f"{os.path.join(path, package)}: not a valid package")
                
        return res
    
    # search package in local project
    if local:
        res += _search_dir('dist')
    if res is not None:
        return res

    # search package in user space
    if _global: 
        res += _search_dir(os.path.join(prefix or "", '~/.army/dist'))

    return res

def load_installed_package(name, local=True, _global=True, prefix=None):
    res = None
    
    def _search_dir(path):
        if os.path.exists(os.path.expanduser(path))==False:
            return None

        found_package = None
        found_version = None
         
        for package in os.listdir(os.path.expanduser(path)):
            if name==package:
                return _load_installed_package(os.path.join(path, package))
    
        return found_package
    
    # search package in local project
    if local:
        res = _search_dir('dist')

    # search package in user space
    if res is None and _global: 
        res = _search_dir(os.path.join(prefix or "", '~/.army/dist'))
    
    return res

# TODO check version when loading package and in case of package installed both global and local use the best fit
def load_project_packages(project, target):
    to_load = []
    for dependency in project.dependencies:
        to_load.append(dependency)

    for dependency in project.target[target].dependencies:
        to_load.append(dependency)

    dependencies = []
    while len(to_load)>0:
        dependency = to_load.pop(0)
        installed = load_installed_package(dependency)
        dependencies.append(installed)
        
        for dependency in installed.dependencies:
            to_load.append(dependency)
    
    return dependencies

def _load_installed_package(path):
    # TODO find a way to add line to error message
    file = os.path.expanduser(os.path.join(path, 'army.toml'))
    if os.path.exists(file)==False:
        raise PackageException(f"{file}: file not found")

    content = {}
    try:
        log.info(f"load installed package '{file}'")
        content = toml.load(file)
        log.debug(f"content: {content}")
    except Exception as e:
        print_stack()
        log.debug(e)
        raise PackageException(f"{format(e)}")
    
    project = InstalledPackage(data=content, path=path)
    project.check()

    return project

    
class PackageException(Exception):
    def __init__(self, message):
        self.message = message

class Package(Schema):
    def __init__(self, data, schema):
        super(Package, self).__init__(data, schema={
                **schema,
                'name': String(),
                'description': String(),
                'version': VersionString(),
                'dependencies': Optional(VariableDict(PackageString(), VersionRangeString())),
                'plugins': Optional(VariableDict(PackageString(), VersionRangeString())),
                'plugin': Optional(VariableDict(PackageString(), Variant())),
                'packaging': Optional(Dict({
                    'include': Optional(Array(String())),
                    'exclude': Optional(Array(String()))
                    })),
                
                # arch definition
                'arch': Optional(VariableDict(String(), Dict({
                    'definition': Optional(String()),
                    }))),

                # in case of a firmware
                'default-target': Optional(String()),
                'target': Optional(VariableDict(String(), Dict({
                    'arch': String(),
                    'definition': Optional(String()),
                    'dependencies': Optional(VariableDict(PackageString(), VersionRangeString())),
                    }))),
                
                # in case of a library
                'cmake': Optional(Dict({
                    'include': Optional(String()),
                    })),

           })
    
    @property
    def name(self):
        return self._data['name']
    
    @property
    def description(self):
        return self._data['description']
    
    @property
    def version(self):
        return Version(self._data['version'])
    # 
    @property
    def arch(self):
        class ArchDictIterator(object):
            def __init__(self, values):
                self._list = values
                self._iter = iter(self._list)
             
            def __next__(self):
                return next(self._iter)

        class ArchDict(object):
            def __init__(self, data):
                self._data = data
                
            def __iter__(self):
                return ArchDictIterator(self._data)
            
            def __getitem__(self, item):
                return Arch(self._data[item])
            
        class Arch(object):
            def __init__(self, data):
                self._data = data
            
            @property
            def definition(self):
                return self._data['definition']
        
        if 'arch' in self._data:
            return ArchDict(self._data['arch'])
        return ArchDict({})

    @property
    def dependencies(self):
        if 'dependencies' in self._data:
            return self._data['dependencies']
        return []
    
    @property
    def plugins(self):
        if 'plugins' in self._data:
            return self._data['plugins']
        return []
    
    @property
    def plugin(self):
        if 'plugin' in self._data:
            return self._data['plugin']
        return []

    @property
    def packaging(self):
        class Packaging(object):
            def __init__(self, data):
                self._data = data
        
            @property
            def include(self):
                if 'include' in self._data:
                    return self._data['include']
                return []
        
            @property
            def exclude(self):
                if 'exclude' in self._data:
                    return self._data['exclude']
                return []

        if 'packaging' in self._data:
            return Packaging(self._data['packaging'])
        return Packaging({})
        
    @property
    def default_target(self):
        if 'default-target' in self._data:
            return self._data['default-target']
        return None
        
    @property
    def target(self):
        class TargetDictIterator(object):
            def __init__(self, values):
                self._list = values
                self._iter = iter(self._list)
             
            def __next__(self):
                return next(self._iter)

        class TargetDict(object):
            def __init__(self, data):
                self._data = data
                
            def __iter__(self):
                return TargetDictIterator(self._data)
            
            def __getitem__(self, item):
                return Target(self._data[item])
            
        class Target(object):
            def __init__(self, data):
                self._data = data
            
            @property
            def arch(self):
                return self._data['arch']
            
            @property
            def definition(self):
                return self._data['definition']

            @property
            def dependencies(self):
                if 'dependencies' in self._data:
                    return self._data['dependencies']
                return []
        
        if 'target' in self._data:
            return TargetDict(self._data['target'])
        return TargetDict({})

    @property
    def cmake(self):
        class CMake(object):
            def __init__(self, data):
                self._data = data
            
            @property
            def include(self):
                if 'include' in self._data:
                    return self._data['include']
                return None
        if 'cmake' in self._data:
            return CMake(self._data['cmake'])
        return CMake({})
        
    def __repr__(self):
        return f"{self.name}:{self.version}"

class InstalledPackage(Package):
    def __init__(self, data, path):
        super(InstalledPackage, self).__init__(data, schema={
                'repository': Variant()
            })

        self._path = path
        
    @property
    def path(self):
        return self._path

    @property
    def repository(self):
        class Repository(object):
            def __init__(self, data):
                self._data = data
        
            @property
            def name(self):
                return self._data['name']
        
            @property
            def uri(self):
                return self._data['uri']

        return Repository(self._data['repository'])

    def uninstall(self):
        def rmtree_error(func, path, exc_info):
            print(exc_info)
            exit(1)
        if os.path.exists(self._path):
            log.debug(f"rm {self._path}")
            shutil.rmtree(self._path, onerror=rmtree_error)
    


from army.army import cli
from army.api.click import verbose_option 
import click

@cli.command(name='goodbye', help='Goodbye world')
@verbose_option()
@click.pass_context
def goodbye(ctx, **kwargs):
    print("Goodbye world !")


from army.army import cli
from army.api.click import verbose_option 
import click

@cli.command(name='hello', help='Hello world')
@verbose_option()
@click.pass_context
def hello(ctx, **kwargs):
    print("Hello world !")


from army.api.log import log
from army.api.debugtools import print_stack
from army.api.project import load_project
from army.army import prefix
from army.api.click import verbose_option 
from army.api.package import load_installed_packages
from army.army import cli, dependencies
import click
import sys
import os

@dependencies.command(name='list', help='List installed packages')
@verbose_option()
@click.pass_context
def list(ctx, **kwargs):
    log.info(f"list")
    
    # load configuration
    config = ctx.parent.config

    packages = load_installed_packages(prefix=prefix)

    if len(packages)==0:
        print('no package found', file=sys.stderr)
        return
 
    column_package = ['package']
    column_version = ['version']
    column_description = ['description']
    column_repo = ['repository']
#
    for package in packages:
        column_package.append(package.name)
        column_version.append(str(package.version))
        column_description.append(package.description)
        column_repo.append(package.repository.name)
  
    max_package = len(max(column_package, key=len))
    max_version = len(max(column_version, key=len))
    max_description = len(max(column_description, key=len))
    max_repo = len(max(column_repo, key=len))
  
    for i in range(len(column_repo)):
        print(f"{column_package[i].ljust(max_package)} | ", end='')
        print(f"{column_version[i].ljust(max_version)} | ", end='')
        print(f"{column_description[i].ljust(max_description)} | ", end='')
        print(f"{column_repo[i].ljust(max_repo, ' ')}", end='')
        print()



from army.api.log import log
from army.api.package import load_installed_package
from army.api.debugtools import print_stack
from army.api.version import Version
import importlib.util
import os

def load_plugin(name, config):
    log.info(f"load plugin '{name}'")

    search_name = name
    search_version = None
    if ':' in name:
        search_name, search_version = name.split(':')

    package = load_installed_package(search_name)

    if package is None:
        raise PluginException(f"{name}: plugin not installed")
        return

    if search_version and package.version!=Version(search_version):
        raise PluginException(f"{name}: require version '{search_version}' but '{package.version}' found")

    try:
        spec = importlib.util.spec_from_file_location("plugin", os.path.join(package.path, '__init__.py'))
        plugin = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(plugin)
    except Exception as e:
        print_stack()
        log.debug(e)
        raise PluginException(f"{name}: failed to load plugin")

class PluginException(Exception):
    def __init__(self, message):
        self.message = message




import toml
import os
import sys
from army.api.log import log
from army.api.debugtools import print_stack
from army.api.version import Version

# load army configuration files, each file supersedes the previous
# Global configuration: /etc/army/army.toml
# User configuration: ~/.army/army.tom
# @param parent parent configuration
# @param prefix mainly used for unit tests purpose
def load_configuration(parent=None, prefix=None):
    # load main config file
    config = ArmyConfig(parent=parent)
    path = os.path.join(prefix or "/", 'etc/army/army.toml')
    if os.path.exists(path):
        config = load_configuration_file(path=path, parent=parent)
    
    # load main repositories
    path = os.path.join(prefix or "/", 'etc/army/repo.d')
    if os.path.exists(path) and os.path.isdir(path):
        for f in os.listdir(path):
            if os.path.isfile(os.path.join(path, f)) and f.endswith(".toml"):
                config = load_configuration_repository_file(path=os.path.join(path, f), parent=config)

    # load user config file
    path = os.path.join(prefix or "", '~/.army/army.toml')
    if os.path.exists(os.path.expanduser(path)):
        config = load_configuration_file(path=path, parent=config)
     
    # load user repositories
    path = os.path.join(prefix or "", '~/.army/repo.d')
    if os.path.exists(os.path.expanduser(path)) and os.path.isdir(os.path.expanduser(path)):
        for f in os.listdir(os.path.expanduser(path)):
            if os.path.isfile(os.path.join(os.path.expanduser(path), f)) and f.endswith(".toml"):
                config = load_configuration_repository_file(path=os.path.join(path, f), parent=config)

    return config


def load_configuration_file(path, parent=None):
    # TODO find a way to add line to error message
    file = os.path.expanduser(path)
    if os.path.exists(file)==False:
        raise ConfigException(f"{file}: file not found")

    config = {}
    try:
        log.info(f"load config '{path}'")
        config = toml.load(file)
        log.debug(f"content: {config}")
    except Exception as e:
        print_stack()
        log.debug(e)
        raise ConfigException(f"{format(e)}")
    
    res = ArmyConfig(parent=parent, value=config)
    return res
#     for item in config:
#         try:
#             log.debug(f"load '{item}': {config[item]}")
#             res.set(item, config[item])
#         except Exception as e:
#             print_stack()
#             log.debug(e)
#             raise ConfigException(f"{path}: {e}")
#     res.check()
    

def load_configuration_repository_file(path, parent=None):
    # TODO find a way to add line to error message
    file = os.path.expanduser(path)
    if os.path.exists(file)==False:
        raise ConfigException(f"{file}: file not found")

    config = {}
    try:
        log.info(f"Load config '{path}'")
        config = toml.load(file)
        log.debug(f"content: {config}")
    except Exception as e:
        print_stack()
        log.debug(e)
        raise ConfigException(f"{format(e)}")
    
    res = ArmyConfigRepository(parent=parent, value=config)
    return res
#     for item in config:
#         try:
#             log.debug(f"load '{item}': {config[item]}")
#             res.set(item, config[item])
#         except Exception as e:
#             print_stack()
#             log.debug(e)
#             raise ConfigException(f"{path}: {e}")
    

    
class ConfigException(Exception):
    def __init__(self, message):
        self.message = message

# Configuration base class
class BaseConfig(object):

    def __init__(self, value=None, parent=None):
        super(BaseConfig, self).__init__()
        
        self._parent = parent
        self._value = value
        
    def parent(self):
        return self._parent 

    def set_parent(self, parent):
        self._parent = parent
        self.check()

    # get item value stored inside the item
    def value(self):
        return self._value


class Config(BaseConfig):
    ITEM_TYPE=0          # mandatory: field type
    ITEM_DEFAULT_VALUE=1 # mandatory: define default field value
    ITEM_ALLOCATOR=2     # facultative: custom allocator for field

    def __init__(self, value=None, parent=None, fields=None):
        super(Config, self).__init__(value=value, parent=parent)
        
        # configuration fields
        # key: field name
        # value: [ <item class>, default value ]
        self._fields = fields
        
        self.load() ;
        
    # check value content
    def check(self):
        # check for mandatory fields
        for field_name in self.get_fields():
            field = self.get_field(field_name)

            if isinstance(field, list)==False:
                raise ConfigException(f"{field_name}: invalid field format")                
            if len(field)==0:
                raise ConfigException(f"{field_name}: missing field type")
            if len(field)==1 and ( self._value is None or field_name not in self._value):
                raise ConfigException(f"{field_name}: missing value")
                
        
        # check for extraneous elements
        if isinstance(self._value, dict):
            err = []
            # check all fields
            for value in self._value:
                if self.get_field(value) is None:
                    raise ConfigException(f"{value}: unexpected value")
#             for value in self._value:
#                 try:
#                     self.get(value)
#                 except ConfigException as e:
#                     err.append(f"{e}")
#                     
#             if len(err)>0:
#                 raise ConfigException(err)

    def load(self):
        self.check() 

    def get_fields(self):
        if self._fields:
            return self._fields
        return {}

    def get_field(self, item):
        if self._fields and item in self._fields:
            return self._fields[item]
            
        if self._parent:
            return self._parent.get_field(item)
        return None
    
    # @description get subitem value
    # @param item name of the item to look for
    # @param default when the item is not found, if True  return its default value, when False return None
    def get(self, item, default=True):
        # check that field exists somewhere in the configuration tree
        field = self.get_field(item)
        
        if field is None:
            raise ConfigException(f"'{item}': unexpected value")

        if len(field)==3:
            # custom allocator given for this field
            allocator = field[self.ITEM_ALLOCATOR]
        else:
            # use default allocator
            allocator = field[self.ITEM_TYPE]
        
        value = None
        if isinstance(self.value(), dict) and item in self.value():
            # item exists in the current configuration
            # instantiate configuration item from loaded value
            value = self.value()[item]
        elif self.parent() and self.parent().get_field(item) is None:
            # item does not exists either in the parent
            if default==True:
                # set default value
                value = field[self.ITEM_DEFAULT_VALUE]
            else:
                # no value
                value = None
        elif self.parent():
            # parent exists and has the item
            value = self.parent().get(item).value()

        if value:
            # we have a value
            return allocator(value=value)
        elif default==True:
            # item not found in herarchy, return default value
            return allocator(value=field[self.ITEM_DEFAULT_VALUE])
        return None

    def __getattribute__(self, name):
        try:
            return object.__getattribute__(self, name)
        except AttributeError:
            if self.get_field(name):
                return object.__getattribute__(self, "get")(name)
            else:
                raise AttributeError(name)
    
    def set(self, item, value):
        field = self.get_field(item)

        if field is None:
            raise ConfigException(f"'{item}': unexpected value")
        
        if self._value is None:
            self._value = {}
            
        if self._value and not isinstance(self._value, dict):
            # this should not happen, means there is a problem to investigate upstream
            raise ConfigException(f"'{item}': parent type mismatch")
        
        # create an item, should trigger an error if content is invalid
        v = field[self.ITEM_TYPE](value=value)
        self._value[item] = v.value()
        
    def expand(self, default=True):
        res = None
        if self._fields:
            res = {}
        else:
            return self._value
        
        if self.parent():
            res = self.parent().expand(default)
        
        for item in self.get_fields():
#            field = self.get_field(item)
            value = self.get(item, default)
            if value:
                res[item] = value.expand()

        return res

    def __str__(self):
        return f"{self.value()}"

class ConfigNullableString(Config):
    def __init__(self, value=None, parent=None):
        super(ConfigNullableString, self).__init__(value=value, parent=parent)
    
    def check(self):
        if self.value() and not isinstance(self.value(), str):
            raise ConfigException(f"'{self.value()}': invalid string")

    
class ConfigString(Config):
    def __init__(self, value=None, parent=None):
        super(ConfigString, self).__init__(value=value, parent=parent)
    
    def check(self):
        if not isinstance(self.value(), str):
            raise ConfigException(f"'{self.value()}': invalid string")

    def __str__(self):
        return f"{self.value()}"
    
    def __repr__(self):
        return f"{self.value()}"

class ConfigVersion(Config):
    def __init__(self, value=None, parent=None):
        super(ConfigVersion, self).__init__(value=Version(value), parent=parent)

    def check(self):
        if not isinstance(self.value(), Version):
            raise ConfigException(f"'{self.value()}': invalid version")

class ConfigInt(Config):
    def __init__(self, value=None, parent=None):
        super(ConfigInt, self).__init__(value=value, parent=parent)

    def check(self):
        if not isinstance(self.value(), int):
            raise ConfigException(f"'{self.value()}': invalid integer")
    
    def __int__(self):
        return self._value


class _ConfigDictIterator(object):
    def __init__(self, values):
        self._list = values
        
        self._items = values.value()
        parent = values.parent()
        while parent:
            self._items = { **parent.value(), **self._items }
            parent = parent.parent()

        self._iter = iter(self._items)
     
    def __next__(self):
        return next(self._iter)
#         return self._list.get_field()[ConfigDict.ITEM_TYPE](value=val)
 

class ConfigDict(BaseConfig):
    ITEM_TYPE=0
    ITEM_DEFAULT_VALUE=1

    def __init__(self, value=None, parent=None, field=None):
        super(ConfigDict, self).__init__(value=value, parent=parent)

        # fields will contain only the type of dict items
        self._field = field

        self.load() ;
 
    def check(self):
        if self._parent and not isinstance(self.parent(), ConfigDict):
            raise ConfigException(f"'{self.parent()}': invalid parent type")
        
        if not isinstance(self.value(), dict):
            raise ConfigException(f"'{self.value()}': invalid value dict")
         
    def load(self):
        self.check()
        
    def get_field(self):
        return self._field
    
    def append(self, name, item):
        # check that field exists somewhere in the configuration tree
        field = self.get_field()
        
        if isinstance(item, type(field))==False:
            raise ConfigException(f"{item}: invalid type")
        
        self.value()[name] = item.value()
        
    def expand(self):
        res = None
        if self._field:
            res = {}
        else:
            return self._value
        
        if self.parent():
            res = self.parent().expand()

        for item in self:
            res[item] = self[item].expand()

        return res
    
    def __len__(self):
        list = self.value()
        parent = self.parent()
        while parent:
            list = { **parent.value(), **list }
            parent = parent.parent()
        return len(list)
    
    def __getitem__(self, name):
        # check that field exists somewhere in the configuration tree
        field = self.get_field()
        
        if field is None:
            raise ConfigException(f"None type found")

        value = None
        if self.value() and name in self.value():
            # item exists in the current configuration
            # instantiate configuration item from loaded value
            return field[self.ITEM_TYPE](value=self.value()[name])
        elif self.parent():
            return self.parent()[name]
        else:
            raise ConfigException(f"{name}: does not exists")
            
        return None

    def __str__(self):
        return f"{self.value()}"

    def __iter__(self):
        return _ConfigDictIterator(self)


class _ConfigListIterator(object):
    def __init__(self, list):
        self._list = list
#         self._field = field
#         self._list = values
        self._index = 0
#         self._iter = iter(values)
    
    def __next__(self):
        if self._index<len(self._list):
            res = self._list[self._index]
            self._index += 1
            return res
        raise StopIteration()
    
class ConfigList(BaseConfig):
    ITEM_TYPE=0
    ITEM_DEFAULT_VALUE=1

    def __init__(self, value=None, parent=None, field=None):
        super(ConfigList, self).__init__(value=value, parent=parent)

        # fields will contain only the type of dict items
        self._field = field

        self.load() ;

    def check(self):
        if self._parent and not isinstance(self.parent(), ConfigList):
            raise ConfigException(f"'{self.parent()}': invalid parent type")
        
        if not isinstance(self.value(), list):
            raise ConfigException(f"'{self.value()}': invalid value list")

    def load(self):
        self.check()
        
    def get_field(self):
        return self._field
    
    def append(self, item):
        # check that field exists somewhere in the configuration tree
        field = self.get_field()
        
        if isinstance(item, type(field))==False:
            raise ConfigException(f"{item}: invilad type")
        
        self.value().append(item.value())
        
    def expand(self, default=True):
        res = None
        if self._field:
            res = []
        else:
            return self._value
        
        for item in self:
            res.append(item.expand())
        return res
    
    def __len__(self):
        res = len(self.value())
        if self.parent():
            res += len(self.parent())
        return res 
    
    def __getitem__(self, index):
        # check that field exists somewhere in the configuration tree
        field = self.get_field()
        
        if field is None:
            raise ConfigException(f"None type found")

        if index<0 or index>=len(self):
            raise ConfigException(f"{index}: out of bounds")
            
        value = None
        if index<len(self.value()):
            # item exists in the current configuration
            # instantiate configuration item from loaded value
            return field[self.ITEM_TYPE](value=self.value()[index])
        elif self.parent():
            pindex = index-len(self.value())
            return self.parent()[pindex]

        return None

    def __str__(self):
        return f"{self.value()}"

    def __iter__(self):
        return _ConfigListIterator(self)

class ConfigStringList(ConfigList):
    
    def __init__(self, parent=None, value=None):
        super(ConfigStringList, self).__init__(
            parent=parent,
            value=value,
            field=[ ConfigString ]
        )

# TODO: create type ConfigEnum as base class
class ConfigLogLevel(ConfigString):
    def __init__(self, value=None, parent=None):
        super(ConfigLogLevel, self).__init__(value=value, parent=parent)

    def check(self):
        super(ConfigLogLevel, self).check() 
        if self._value not in {"debug", "info", "warning", "error", "fatal"}:
            raise ConfigException(f"'{self.value()}': invalid value")


class ArmyConfig(Config):
    def __init__(self, value=None, parent=None):
        super(ArmyConfig, self).__init__(
            value=value,
            parent=parent,
            fields={
                'verbose': [ ConfigLogLevel, "error" ],
                'target': [ ConfigString, "" ],
                "repo": [ ConfigRepositoryDict, {}, self._allocate_repo ]
            }
        )

    # define a custom allocator for the repo field, this field is not replaced by childs but needs to be merged
    # as each child just adds or superseed repository list
    def _allocate_repo(self, value):
        res = ConfigRepositoryDict(value=value)
        current = res
        
        # merge repository list with parent repo
        parent = self.parent()
        while(parent):
            try:
                if(parent.get_field('repo')):
                    new = ConfigRepositoryDict(value=parent.repo.value())
                    current.set_parent(new)
                    current = new
            except:
                pass
            parent = parent.parent()

        if parent:  
            return ConfigRepositoryDict(value=value, parent=parent.repo)
        return res
        

class ConfigRepositoryDict(ConfigDict):
    
    def __init__(self, parent=None, value=None):
        super(ConfigRepositoryDict, self).__init__(
            parent=parent,
            value=value,
            field=[ ConfigRepository, {} ]
        )
        
class ConfigRepository(Config):
    def __init__(self, parent=None, value=None):
        super(ConfigRepository, self).__init__(
            parent=parent,
            value=value,
            fields={
                'type': [ ConfigString, "" ],
                'uri': [ ConfigString, "" ],
            }
        )
    
class ArmyConfigRepository(Config):
    def __init__(self, parent=None, value=None):
        super(ArmyConfigRepository, self).__init__(
            parent=parent,
            value=value,
            fields={
                'repo': [ ConfigRepositoryDict, {}, self._allocate_repo ]
            }
        )

    # define a custom allocator for the repo field, this field is not replaced by childs but needs to be merged
    # as each child just adds or superseed repository list
    def _allocate_repo(self, value):
        res = ConfigRepositoryDict(value=value)
        current = res
        
        # merge repository list with parent repo
        parent = self.parent()
        while(parent):
            try:
                if(parent.get_field('repo')):
                    new = ConfigRepositoryDict(value=parent.repo.value())
                    current.set_parent(new)
                    current = new
            except:
                pass
            parent = parent.parent()

        if parent:  
            return ConfigRepositoryDict(value=value, parent=parent.repo)
        return res
    




from army.api.log import log
from army.api.package import Package, PackageException
from army.api.debugtools import print_stack
from army.api.version import Version, VersionRange
import os
import subprocess
import shutil
import toml

repository_types = {}

def register_repository(repository_class):
    global repository_types
    
    if repository_class.TYPE:
        repository_types[repository_class.TYPE] = repository_class
        log.debug(f"registered '{repository_class.TYPE}' repository type")


# build repository list from configuration
def load_repositories(config, prefix=None):
    global repository_types
    res = []
    
    if prefix is None:
        log.debug(f"load repositories")
    else:
        log.debug(f"load repositories from {prefix}")

    repos = {}
    try:
        repos = config.repo
    except Exception as e:
        print_stack()
        log.debug(f"{e}")
        log.warning("No repository found")
    
    for repo_name in repos:
        try:
            repo_type_name = repos[repo_name].type.value()
            repo_uri = os.path.join(prefix or "", repos[repo_name].uri.value())

            if repo_type_name in repository_types:
                    
                repo_type = repository_types[repo_type_name]
                
                # instanciate repository and load it
                repo = repo_type(name=repo_name, path=repo_uri)
                res.append(repo)
            else:
                log.warning(f"{repo_type_name}: unhandheld repository type")
        except Exception as e:
            print_stack()
            log.debug(f"{e}")
            log.error(f"{repo_name}: load repository failed")
            
    return res

class RepositoryException(Exception):
    def __init__(self, message):
        self.message = message

# TODO: implement source
class Repository(object):
    TYPE=None
    DEV=False
    
    def __init__(self, name, uri):
        self._name = name
        self._uri = uri
    
    @property
    def uri(self):
        return self._uri
    
    @property
    def name(self):
        return self._name
    
    # override to return package list
    @property
    def packages(self):
        return [] 
    
    @property
    def type(self):
        return self.TYPE
    
    # load package list from repository
    def load(self):
        pass

    # if repository has an internal cache this is where it should be reconstructed
    def update(self):
        self.load()
    
    # search for a package inside the package list
    # @param fullname if True then package name must match exactly, if version is given then fullname is True
    def search(self, name, version=None, fullname=False):
        versions = {}
        
        packages = self.packages

        if version is not None:
            fullname = True
        
        # select packages matching name criteria in package list
        for package in packages:
            match_name = False
            
            if fullname==False and name in package.description.lower():
                match_name = True
                
            if fullname==True and name==package.name:
                match_name = True
            elif fullname==False and name in package.name:
                match_name = True
            
            if match_name==True:
                if package.name in versions:
                    versions[package.name].version.add_version(package.version)
                else:
                    if version is None:
                        # no version specified, give latest by default
                        versions[package.name] = VersionRange(value='latest', versions=[package.version])
                    else:
                        versions[package.name] = VersionRange(value=version, versions=[package.version])

        res = {}
        # select packages matching version in found packages
        for name in versions:
            for package in packages:
                if package.version==versions[name].value:
                    res[name] = package
                    
        return res


class RepositoryPackage(Package):
    def __init__(self, data, repository):
        super(RepositoryPackage, self).__init__(data=data, schema={})
        self._repository = repository
    
    @property
    def repository(self):
        return self._repository
        
    def install(self, path, link):
        includes = self.packaging.include
        
        dest = path

        def rmtree_error(func, path, exc_info):
            print(exc_info)
            exit(1)
        if os.path.exists(path):
            log.debug(f"rm {path}")
            shutil.rmtree(path, onerror=rmtree_error)

        # execute preinstall step
        if os.path.exists(os.path.expanduser(os.path.join(self._repository._uri, 'pkg', 'preinstall'))):
            log.info("execute preinstall script")
            subprocess.check_call([os.path.join(os.path.expanduser(self._repository._uri), 'pkg', 'preinstall')])

        # check that all files exists
        for include in includes:
            source = os.path.join(self._repository._uri, include)
            if os.path.exists(os.path.expanduser(source))==False:
                raise PackageException(f"{include}: package include file not found")

        # create detination tree
        if os.path.exists(dest)==False:
            os.makedirs(dest)
            
        for include in includes:
            source = os.path.join(self._repository._uri, include)
            
            if link==True:
                self._link(source, dest)
            else:
                self._copy(source, dest)
        
        # add repository informations to army.toml
        try:
            self._copy(os.path.join(self._repository._uri, "army.toml"), dest)
            filepath = os.path.join(dest, 'army.toml')
            content = toml.load(filepath)
            content['repository'] = {
                'uri': self._repository._uri,
                'name': self._repository._name
                }
            with open(filepath, "w") as file:
                toml.dump(content, file)
        except Exception as e:
            print_stack()
            log.debug(e)

        #execute postinstall command
        if os.path.exists(os.path.expanduser(os.path.join(self._repository._uri, 'pkg', 'postinstall'))):
            log.info("execute postinstall script")
            subprocess.check_call([os.path.join(os.path.expanduser(self._repository._uri), 'pkg', 'postinstall')])
        
    def _copy(self, source, dest):
        log.debug(f"copy {source} -> {dest}")
        source = os.path.expanduser(source)
        if os.path.isfile(source):
            shutil.copy(source, os.path.join(dest, os.path.basename(source)))
        else:
            shutil.copytree(source, os.path.join(dest, os.path.basename(source)))

    def _link(self, source, dest):
        log.debug(f"link {source} -> {dest}")
        source = os.path.expanduser(source)
        os.symlink(os.path.abspath(source), os.path.join(dest, os.path.basename(source)))



from army.api.repository import Repository
from army.api.log import log

class GitRepositoryException(Exception):
    def __init__(self, message):
        self.message = message

class GitRepository(Repository):
    TYPE="git"
    DEV=False
    
    def __init__(self, name, path):
        super(GitRepository, self).__init__(name=name, uri=path)
        
        self._packages = []
        
        self.load()

    # load package list from repository
    def load(self):
        raise GitRepositoryException(f"{self.name()}: GitRepository: not yet implemented")

    def packages(self):
        return self._packages 



from army.api.log import log
from army.api.debugtools import print_stack
from army.api.project import load_project
from army.api.repository import load_repositories
from army.army import prefix
from army.api.click import verbose_option 
from army.api.package import load_installed_packages, load_installed_package
from army.army import cli, dependencies
import click
import os
import sys

class PackageDependency(object):
    def __init__(self, package, repository, from_package=None):
        self._package = package
        self._from_package = from_package
        self._repository = repository
        
    @property 
    def package(self):
        return self._package
    
    @property
    def from_package(self):
        return self._from_package

    @property
    def repository(self):
        return self._repository

    def __repr__(self):
        return f"{self._package.name}:{self._package.version}"

@dependencies.command(name='install', help='Install package')
@verbose_option()
@click.option('-l', '--link', help='Link files instead of copy (local repository only)', is_flag=True)
@click.option('-g', '--global', help='Install module in user space', is_flag=True)
@click.option('-r', '--reinstall', help='Force reinstall module if already exists', is_flag=True)
# @click.option('--save', help='Update project package list', is_flag=True)    # TODO
@click.argument('name', nargs=-1)
@click.pass_context
def install(ctx, name, link, reinstall, **kwargs):
    log.info(f"install {name} {kwargs}")
    
    _global = kwargs['global']  # not in parameters due to conflict with global keywoard
    
    # load configuration
    config = ctx.parent.config
    project = None
    if os.path.exists('army.toml'):
        try:
            # load project configuration
            project = load_project()
        except Exception as e:
            print_stack()
            log.debug(e)
    if project is None:
        log.info(f"no project loaded")
            
    if len(name)==0 and project is None:
        print("nothing to install", file=sys.stderr)
        exit(1)
        
    # build repositories list
    repositories = load_repositories(config, prefix)
    
    packages = []

    if len(name)==0:
        if project is None:
            log.error(f"{os.getcwd()}: army.toml not found")
            exit(1)

        for package in project.dependencies:
            pkg, repo = _find_package(package, project.dependencies[package], repositories)
            packages.append(PackageDependency(package=pkg, repository=repo))
            
        for plugin in project.plugins:
            pkg, repo = _find_package(plugin, project.plugins[plugin], repositories)
            packages.append(PackageDependency(package=pkg, repository=repo))
    else:
        for package in name:
            pkg, repo = _find_package(package, 'latest', repositories)
            packages.append(PackageDependency(package=pkg, repository=repo))

    # locate install folder
    if _global:
        path = os.path.join(prefix or "", "~/.army/dist/")
    else:
        path = "dist"
    
    force = False
    if reinstall:
        force = True

    dependencies = []
    while(len(packages)>0):
        # get dependencies from top level package to end level
        package_dep = packages.pop(0)
        package = package_dep.package

        # dependency treated ok, append to list
        dependencies.append(package_dep)
        
        # append dependencies to list
        for dependency in package.dependencies:
            pkg, repo = _find_package(dependency, package.dependencies[dependency], repositories)
            dep_pkg = PackageDependency(package=pkg, repository=repo, from_package=package)
            packages.append(dep_pkg)

        # append plugins to list
        for plugin in package.plugins:
            pkg, repo = _find_package(plugin, package.plugins[plugin], repositories)
            dep_pkg = PackageDependency(package=pkg, repository=repo, from_package=package)
            packages.append(dep_pkg)

    # treat dependencies first
    dependencies.reverse()
    
    log.debug(f"packages: {dependencies}")
    # checks
    _check_dependency_version_conflict(dependencies)
    _check_installed_version_conflict(dependencies)
    
    # install
    for dependency in dependencies:
        install = False
        installed_package = load_installed_package(dependency.package.name, prefix=prefix)
        if installed_package:
            if force==True:
                print(f"reinstall {dependency.package}")
                install = True
            else:
                print(f"package {dependency.package} already installed")
                install = False
        else:
            install = True
            print(f"install package {dependency.package}")
            
        if install==True:
            if dependency.repository.DEV==True:
                dependency.package.install(path=os.path.join(path, dependency.package.name), link=link)
            else:
                # link mode is only possible with repository DEV
                dependency.package.install(path=os.path.join(path, dependency.package.name), link=False)
    
    # TODO save and save-dev

def _check_dependency_version_conflict(dependencies):
    """ Check if dependencies contains same package with version mismatch
    """
    for dependency in dependencies:
        for dep in dependencies:
            if dep.package.name==dependency.package.name and dep.package.version!=dependency.package.version:
                msg = f"'{dependency.package.name}': version conflicts: "
                if dependency.from_package is None:
                    msg += f"'{dependency.package.version}' from project"
                else:
                    msg += f"'{dependency.package.version}' from '{dependency.from_package.name}'"
                msg += " conflicts with "
                if dep.from_package is None:
                    msg += f"'{dep.package.version}' from project"
                else:
                    msg += f"'{dep.package.version}' from '{dep.from_package.name}'"
                print(msg, file=sys.stderr)
                exit(1)

def _check_installed_version_conflict(dependencies):
    """ Check if dependencies contains a package already installed with a version mismatch
    """
    installed = load_installed_packages(prefix=prefix)
    # TODO
    
def _find_package(name, version, repositories):
    # search for module in repositories
    for repo in repositories:
        res = repo.search(name, version, fullname=True)
        if len(res)>0:
            # result can contain only one element as fullname is True
            for package in res:
                return res[package], repo
    print(f"{name}: package not found", file=sys.stderr)
    exit(1)
    



class VersionException(Exception):
    def __init__(self, message):
        self.message = message


class Version():
    
    def __init__(self, value):
        if isinstance(value, Version):
            self._major = value._major
            self._minor = value._minor
            self._revision = value._revision
            self._build = value._build
            self._dev = value._dev
        else:
            self._major = None
            self._minor = None
            self._revision = None
            self._build = None
            self._dev = False
            
            version = value

            # extract build number
            if '-' in value:
                s = value.split('-', 1)
                self._build = s[1]
                value = s[0]
            
            if '.' in value:
                s = value.split('.')
                
                if(len(s)>3):
                    raise VersionException(f"Invalid version {version}")
                    
                try:
                    if len(s)==3:
                        self._major = int(s[0])
                        self._minor = int(s[1])
                        self._revision = int(s[2])
                    elif len(s)==2:
                        self._major = int(s[0])
                        self._minor = int(s[1])
                except Exception as e:
                    raise VersionException(f"Invalid version {version}")
                    
            else:
                try:
                    self._major = int(value)
                except Exception as e:
                    raise VersionException(f"Invalid version {version}")
            
            if self._build and ( self._major is None or self._minor is None or self._revision is None):
                raise VersionException(f"Invalid version {version}")
            elif self._build and self._build=="dev":
                self._dev = True
        
    def compare(self, version):
        l = [self._major, self._minor, self._revision, self._build]
        r = [version._major, version._minor, version._revision, version._build]
        
        if self._dev and version._dev:
            return 0
        elif self._dev and not version._dev:
            return -1
        elif not self._dev and version._dev:
            return 1
        
        res = 0
        for i in range(3):
            if res==0:
                if l[i] is None or r[i] is None:
                    res = 0
                elif l[i] is not None and r[i] is not None:
                    if l[i]<r[i]:
                        res = -1
                    if l[i]>r[i]:
                        res = 1
                
        if l[3] is not None and r[3] is None:
            return -1
        if l[3] is None and r[3] is not None:
            return 1
        if l[3] is not None and r[3] is not None:
            if l[3]<r[3]:
                return -1
            elif l[3]==r[3]:
                return 0
            elif l[3]>r[3]:
                return 1

        return res
    
    def __str__(self):
        version = f"{self._major}"
        if self._minor is not None:
            version += f".{self._minor}"
        if self._revision is not None:
            version += f".{self._revision}"
        if self._build is not None:
            version += f"-{self._build}"
        return version
    
    def __repr__(self):
        return self.__str__()
    
    def __eq__(self, other):
        return self.compare(other)==0
    
    def __lt__(self, other):
        return self.compare(other)<0

    def __le__(self, other):
        return self.compare(other)<=0

class VersionRange(object):
    def __init__(self, value, versions={}):
        self._tree = {} # for future parsing 
        self._value = value
        self._versions = versions

        self.parse(value)
    
        
    def parse(self, value):
        if value=='latest':
            return
        else:
            Version(value)
    
    @property
    def value(self):
        if self._value=='latest':
            if len(self._versions)==0:
                raise VersionException(f"no version range specified")
            max = Version(self._versions[0])
            for version in self._versions:
                if Version(version)>max:
                    max = Version(version)
            return max
        else:
            return Version(self._value)
    
    def add_version(self, value):
        self._versions.append(value)
    
    def match(self, value):
        if Version(value)==self.value():
            return True
        else:
            for version in self._versions:
                if Version(version)==Version(value):
                    return True
        return False
    
    def __str__(self):
        return str(self.value())
    
    def __repr__(self):
        return self.__str__()



# click extentions and helpers
from army.api.log import set_log_level

import click
from army.api.click.core import Section

def verbose_option(*param_decls, **attrs):
    """Adds a ``-v`` option to a command
    option is handled directyl and sets the logger to the right level
    """

    def decorator(f):
        def callback(ctx, param, value):
            verbose = None
            
            # init verbose level from parent config if exists
            if ctx.parent and ctx.parent.config:
                verbose = ctx.parent.config.verbose
            
            set_log_level(verbose)
            
            if value:
                if len(value)==1:
                    set_log_level('error')
                elif len(value)==2:
                    set_log_level('warning')
                elif len(value)==3:
                    set_log_level('info')
                else:
                    set_log_level('debug')

        attrs.setdefault("is_flag", True)
        attrs.setdefault("multiple", True)
        attrs.setdefault("help", "Activate verbose/debug mode")
        attrs["callback"] = callback
        return click.option(*(param_decls or ("-v",)), **attrs)(f)

    return decorator

def section(name=None, **attrs):
    """Creates a new :class:`Section` with a function as callback.  This
    works otherwise the same as :func:`command` just that the `cls`
    parameter is set to :class:`Group`.
    """
    attrs.setdefault("cls", Section)
    return click.command(name, **attrs)


def _check_section_params(cmd):
    if not isinstance(cmd, Section) or len(cmd.params)==0:
        return
    
    raise RuntimeError(
        f"Section '{cmd.name}' can not have any parameter as they would be ignored."
    )



import sys, traceback
from army.api.log import log

def print_stack():
    if log.level==10:
#        traceback.print_stack()
        exc_type, exc_value, exc_traceback = sys.exc_info()
        traceback.print_tb(exc_traceback, file=sys.stderr, limit=None)

def print_callstack():
    if log.level==10:
        for line in traceback.format_stack()[:-2]:
            print(line.strip())



#!/usr/bin/env python
# coding=utf-8

#
# To publish package release
# python setup.py sdist upload -r pypi
#

from setuptools import setup

setup(
    name='army',
    version='0.1.0',
    description='Arm cross compiling toolset',
    url='https://github.com/turdusmerula/army/',
    author='Sebastien Besombes',
    license='GPLv3',

#     scripts=['army.py'],
    packages=['army'],
#     package_data={
#         '' : ['*'],
#         'army.library-template': ['*'], 
#         'plugin-template': ['*'], 
#         'army.project-template': ['*'], 
#         },
    include_package_data=True,
    entry_points={
        'console_scripts': [
            'army = army.army:main',
        ]
    },
    zip_safe=False, # force egg extraction
    
    install_requires=[
        'tornado',      # template manager
        'gitpython',    # git repository manager
        'toml',         # toml files
        'click',
        'cmake',
        'schema',
    ],
)




import sys
import os
#sys.path.append(os.path.dirname(__name__))

from army.api.version import Version

print(Version("1"))
print(Version("1.2"))
print(Version("1.2.3"))
print(Version("1.2.3-b"))
print(Version("1.2.3-dev"))

# validated with https://semver.npmjs.com/

values = [
    "",
    "1.1.1.1",
    "1.a",
    "1-a"
]

for value in values:
    try:
        print(f"version: {value}")
        Version(value)
        assert False
    except AssertionError:
        print("Should assert !")
        exit(1)
    except Exception as e:
        print(format(e))
    print("---")

values = [
    #           [L<R    L<=R   L>R    L>=R   L==R ]
    ["1", "1",  [False, True,  False, True,  True ]],
    ["1", "2",  [True,  True,  False, False, False]],
    ["2", "1",  [False, False, True,  True,  False]],
    
    ["2.0", "2.0",  [False, True,  False, True,  True ]],
    ["2.0", "1.9",  [False, False, True,  True,  False]],
    ["1.9", "2.0",  [True,  True,  False, False, False]],

    ["2.0.0", "2.0.0",  [False, True,  False, True,  True ]],
    ["2.0.0", "1.9.9",  [False, False, True,  True,  False]],
    ["1.9.9", "2.0.0",  [True,  True,  False, False, False]],

    ["2.0.0-aa", "2.0.0-aa",  [False, True,  False, True,  True ]],
    ["2.0.0-aa", "2.0.0-ab",  [True,  True,  False, False, False]],
    ["2.0.0-ab", "2.0.0-aa",  [False, False, True,  True,  False]],

    ["1.1.1", "1",       [False, True, False,  True,  True]],
    ["1.1.1", "1.1",     [False, True, False,  True,  True]],
    ["1.1.1-rc1", "1.1", [True,  True, False, False, False]],

    ["1.1.1", "1.1.1-dev",       [False, False, True,  True,  False]],
    ["1.1.1-dev", "1.1.1-dev",   [False, True,  False,  True,  True]],
    
    ["2.1.2", "2.1.2",  [False, True,  False, True,  True ]],
    ["2.1.2", "2.1.1",  [False, False, True,  True,  False]],
    ["2.1.2", "2.1.3",  [True,  True,  False, False, False]],
]

for v in values:
    vl = Version(v[0])
    vr = Version(v[1])
    asserts = v[2]
    
    print(f"{vl}<{vr}: {vl<vr}")
    assert (vl<vr)==asserts[0]
    
    print(f"{vl}<={vr}: {vl<=vr}")
    assert (vl<=vr)==asserts[1]
    
    print(f"{vl}>{vr}: {vl>vr}")
    assert (vl>vr)==asserts[2]    
    
    print(f"{vl}>={vr}: {vl>=vr}")
    assert (vl>=vr)==asserts[3]    

    print(f"{vl}=={vr}: {vl==vr}")
    assert (vl==vr)==asserts[4]    

    print("---")




from army.api.repository import Repository, RepositoryPackage
from army.api.project import load_project, Project
from army.api.debugtools import print_stack
from army.api.log import log
from army.api.package import Package
from army.army import prefix
import os
import toml

class LocalGitRepositoryException(Exception):
    def __init__(self, message):
        self.message = message

class LocalGitRepository(Repository):
    TYPE="git-local"
    DEV=True
    
    def __init__(self, name, path):
        super(LocalGitRepository, self).__init__(name=name, uri=path)

        self._project = None
        
        self.load()
        
    # load package list from repository
    def load(self):
        # load project file
        # TODO find a way to add line to error message
        file = os.path.expanduser(os.path.join(prefix or "", self.uri, 'army.toml'))
        if os.path.exists(file)==False:
            raise LocalGitRepositoryException(f"{file}: file not found")
    
        content = {}
        try:
            log.info(f"Load git repository '{file}'")
            content = toml.load(file)
            log.debug(f"content: {content}")
        except Exception as e:
            print_stack()
            log.debug(e)
            raise LocalGitRepositoryException(f"{format(e)}")
        
        self._project = RepositoryPackage(data=content, repository=self)
        self._project.check()
        
    @property
    def packages(self):
        return [self._project] 

    def update(self):
        # nothing to do in a local repository
        pass



from army.api.log import log
from army.api.debugtools import print_stack
from army.api.package import Package
from army.api.schema import SchemaException
import toml
import toml.decoder
import os
import sys

# from army.api.config import ArmyConfigFile, Config, ConfigString, ConfigVersion
# from army.api.config import ConfigList, ConfigStringList, ConfigDict

# load project army file 
# @param parent parent configuration
# @param prefix mainly used for unit tests purpose
# @return the loaded project configuration or None if project was not loaded
def load_project(path='army.toml'):
    
    # TODO find a way to add line to error message
    file = os.path.expanduser(path)
    if os.path.exists(file)==False:
        raise ProjectException(f"{file}: file not found")

    content = {}
    try:
        log.info(f"Load project '{file}'")
        content = toml.load(file)
        log.debug(f"content: {content}")
    except toml.decoder.TomlDecodeError as e:
        print_stack()
        log.debug(e)        
        raise e
    except Exception as e:
        print_stack()
        log.debug(e)
        raise ProjectException(f"{format(e)}")
    
    project = Project(data=content)
    project.check()
        

    return project

class ProjectException(Exception):
    def __init__(self, message):
        self.message = message


class Project(Package):
    def __init__(self, data):
        super(Project, self).__init__(data, schema={})



from army.api.log import log
from army.api.debugtools import print_stack
from army.api.project import load_project
from army.api.repository import load_repositories
from army.army import cli, dependencies
from army.api.click import verbose_option 
import click
import os
from army.army import prefix

# TODO: implement multiple search criteria

@dependencies.command(name='search', help='Search package in repositories')
@verbose_option()
@click.argument('name')
@click.pass_context
def search(ctx, name, **kwargs):
    log.info(f"search {name}")
    
    # load configuration
    config = ctx.parent.config
    
    # build repositories list
    repositories = load_repositories(config, prefix)
    packages = []
     
    for r in repositories:
        res = r.search(name)
        if len(res)>0:
            for pkg in res:
                packages.append(res[pkg])

    if len(packages)==0:
        print(f'No matches found for "{name}"')
        return
 
    column_repo = ['repository']
    column_package = ['package']
    column_version = ['version']
    column_description = ['description']
#
    for package in packages:
        column_repo.append(package.repository.name)
        column_package.append(package.name)
        column_version.append(str(package.version))
        column_description.append(package.description)
  
    max_repo = len(max(column_repo, key=len))
    max_package = len(max(column_package, key=len))
    max_version = len(max(column_version, key=len))
    max_description = len(max(column_description, key=len))
  
    for i in range(len(column_repo)):
        print(f"{column_repo[i].ljust(max_repo, ' ')} | ", end='')
        print(f"{column_package[i].ljust(max_package)} | ", end='')
        print(f"{column_version[i].ljust(max_version)} | ", end='')
        print(f"{column_description[i].ljust(max_version)}", end='')
        print()


class ArchException(Exception):
    def __init__(self, message):
        self.message = message

class Arch(object):
    def __init__(self, name, parent=None):
        self._name = name
        self._parent = parent
    
    def name(self):
        return self._name 
    
    def parent(self):
        return self._parent



